from . import v1
from . import v2
from ._server import run_simple

VERSIONS = {
    "v1": v1,
    "v2": v2,
}

def get_server(version, **kwargs):
    app = VERSIONS[version].app.make_app(**kwargs)
    return app

def default_server():
    return get_server("v1")




import param
import panel as pn
import holoviews as hv
import numpy as np
import base64
import pandas as pd
from pandas.api.types import is_numeric_dtype

from xepmts.web_client.src.shared.templates import ListTemplate
from xepmts.web_client.src.shared import config, get_db
from xepmts.web_client.src.shared._menu import session_menu
from xepmts.web_client.src.shared._session import login_required
from xepmts.web_client.src.shared.models import BaseSection, FilterTool, CenterColumn

DEFAULT_PLOT_NAMES = ["hist", "bar", "heatmap", "scatter", "bivariate",
 "hexbins", "step", "violin", "table", "line", "kde", "box"]
class ResourcePlots(param.Parameterized):
    
    filter_tool = param.ClassSelector(FilterTool)
    plot_name = param.Selector(DEFAULT_PLOT_NAMES)
    partition_size = param.Integer(5000)
    _plot = param.Parameter()
    overrides = param.Dict({})
    loading = param.Boolean(False)
    load_plot = param.Action(lambda self: self.update_plot())

    def _init_view(self):
        if self.filter_tool is None:
            return
        
    def __init__(self, **params):
        super().__init__(**params)
        if self.filter_tool is None:
            return
        def callback(event):
            self._update_plot_options()
        self.filter_tool.param.watch(callback, "resource",)    
        self._update_plot_options()

    @param.depends("loading")
    def view(self):
        if self.loading:
            return CenterColumn(pn.indicators.LoadingSpinner(value=self.loading),
                                sizing_mode="stretch_width")
        else:
            return pn.panel(self.plot_view)

    @param.depends("_plot")
    def plot_view(self, *events):
        if self._plot is None:
            return pn.Column("# Plot will load here.")
        else:
            return self._plot

    def update_plot(self):
        print(self.plot_name)
        if self.filter_tool is None:
            return
        if self.filter_tool.resource is None:
            return
        if not self.plot_name:
            return
        self.loading = True
        try:
            resource = self.filter_tool.get_filtered_resource()
            plotter = resource.paginate(self.partition_size).plot
            plot = getattr(plotter, self.plot_name)(**self.overrides)
            self._plot = pn.panel(plot, sizing_mode="stretch_both")
        finally:
            self.loading = False

    def _update_plot_options(self):
        if self.filter_tool.resource is None:
            return
        names = self.filter_tool.resource.plots + DEFAULT_PLOT_NAMES
        self.param.plot_name.names = {k:k for k in names}
        self.param.plot_name.objects = names
        if self.plot_name not in names and names:
            self.plot_name = names[0]

    def plot_selection_view(self):
        return pn.Column(
            self.param.plot_name,
            self.param.overrides,
            self.param.load_plot,
        )

    def settings(self):
        parameters = ["partition_size"]
        widgets = {"load_plot": pn.widgets.Button, "button_type": "primary"}
        plotting_settings = pn.Param(self.param,
                        parameters=parameters,
                        widgets=widgets,
                        show_name=False)
        return pn.Column(
                    "### Resource settings",
                    self.filter_tool.resource_settings,
                    plotting_settings,
                    pn.layout.Divider(margin=1),
                    "### PMT plot",
                    self.filter_tool.plot_settings,
                    width=270, sizing_mode="fixed")

    @property
    def sections(self):
        s = self.filter_tool.sections
        s["Plot selection"] = self.plot_selection_view
        s["Plot view"] = self.view
        return s

@login_required(pn.state)          
def view(db):
    # db = pn.state.as_cached(pn.state.curdoc.session_context.id, get_db)
    menu = session_menu(pn.state.curdoc.session_context.id)
    template = ListTemplate(
        title="Resource level plots",
        sidebar_footer=menu,
    )
    pn.state.curdoc.theme = template.theme.bokeh_theme

    browser = ResourcePlots(filter_tool=FilterTool(client=db))

    hv.renderer('bokeh').theme = template.theme.bokeh_theme
    template.main[:] = [pn.Card(v, header="## "+k ,sizing_mode="stretch_both") for k,v in browser.sections.items()]
    template.sidebar[:] = [pn.panel(browser.settings)]
    return template


if __name__.startswith("bokeh"):
    view().servable()



import pathlib
import sqlite3 as sql
from os.path import dirname, join

import numpy as np
import pandas as pd
import panel as pn
import holoviews as hv
import param
from datetime import datetime
import time

from xepmts.web_client.src.shared.templates import ListTemplate
from xepmts.web_client.src.shared import config, get_db
from xepmts.web_client.src.shared._menu import session_menu
from xepmts.web_client.src.shared.models import BaseSection, DataLoader, CenterColumn
from xepmts.web_client.src.shared._session import login_required              

from pandas_profiling import ProfileReport

PARENT_DIR = pathlib.Path(__file__).parent
CONFIG_FILE = PARENT_DIR / "configuration.yaml"

class PandasProfiling(BaseSection):
    title = param.String("Collection profile")
    minimal = param.Boolean(False)
    # _report = param.Parameter(precedence=-1)
    update_report_action = param.Action(lambda self: self._update_view(), label="Create plot")

    def _init_view(self):
        self._view = pn.Column("Data not loaded yet.", sizing_mode="stretch_both")
        
    def _update_view(self, *events):
        if self.data is None:
            return
        else:
            report = ProfileReport(self.data,
                                config_file=CONFIG_FILE,
                                title=self.title)
            self._view = pn.Column(
                    pn.pane.HTML(report.to_html(), sizing_mode="stretch_both"),
                    sizing_mode="stretch_both"
                     )

    def settings(self):
        settings_parameters = ["minimal", "title", "update_report_action"]
        return pn.Param(self.param, parameters=settings_parameters)

@login_required(pn.state)                                 
def view(db):
    menu = session_menu(pn.state.curdoc.session_context.id)
    template = ListTemplate(
        title="Pandas profiling",
        sidebar_footer=menu,
    )
    pn.state.curdoc.theme = template.theme.bokeh_theme

    plotter = PandasProfiling()
    loader = DataLoader(client=db, data_plotter=plotter, _logged_in=db.logged_in)
    loader.merge_pmt_info = False
    hv.renderer('bokeh').theme = template.theme.bokeh_theme
    cards = loader.cards()
    panels = cards[:-1] + [pn.panel(plotter.view)]
    template.main[:] = panels
    template.sidebar[:] = [pn.panel(loader.settings), pn.layout.Divider()]

    return template

if __name__.startswith("bokeh"):
    view().servable()



"""Console script for xepmts."""
import sys
import os

import click
import xepmts


@click.group()
def main():
    """Console script for xepmts."""
    return 0


@main.group()
def server():
    pass

@server.command()
@click.option('--address', default="localhost", help='Server address.')
@click.option('--port', default=5006, help='Server port.')
@click.option('--debug', default=False, help='Enable debugging.', is_flag=True)
@click.option('--reload', default=False, help='Enable auto-reload on code change.', is_flag=True)
@click.option('--evalex', default=False, help='Enable Evalex.', is_flag=True)
def serve(address, port, debug, reload, evalex):
    from xepmts.api.server import run_simple
    run_simple(address, port,  debug, reload, evalex)
   
@main.group()
def webclient():
    pass

@webclient.command()
@click.option('--address', default="localhost", help='Server address.')
@click.option('--port', default=5006, help='Server port.')
@click.option('--nproc', default=1, help='Number of Processes to spawn.')
@click.option('--allow_websocket_origin', default=["localhost"],  multiple=True, help='Whitelisted origins for ws connections.')
@click.option('--debug', default=False, help='Enable auto-reload on code change.', is_flag=True)
def serve(address, port, nproc,
          allow_websocket_origin, debug):
    origins = []
    for origin in allow_websocket_origin:
        origin = origin.strip("/")
        # if not origin.startswith("http"):
        #     origin = "http://" + origin
        if not origin.endswith(str(port)):
            origin = origin + f":{port}"
        origins.append(origin)
    from xepmts.web_client.src import site
    site.serve(address=address, port=port, nproc=nproc,
              allow_websocket_origin=origins, debug=debug)
    


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



"""Provides

- config: An instance of SiteConfig holding the current Site configuration
"""
import os
import pathlib

from xepmts.web_client.src.shared._config import _create_site_config
from xepmts.web_client.src.shared._db import get_db


_SITE_CONFIG_STR = os.getenv("SITE_CONFIG")
if _SITE_CONFIG_STR:
    config = _create_site_config(config_root=_SITE_CONFIG_STR)
else:
    config = _create_site_config(config_root=pathlib.Path(__file__).parent.parent / "config")



"""Test of the App"""
# pylint: disable=redefined-outer-name,protected-access
# pylint: disable=missing-function-docstring,missing-module-docstring,missing-class-docstring
from .app import view


def test_view():
    view()



# -*- coding: utf-8 -*-
import os
# from xepmts.api.server.v1.settings import get_settings_dict

from xepmts.api.server.v1 import settings as settings
settings_file = settings.__file__

def make_app(settings=settings_file, auth=None, app=None,
             swagger=False, fs_store=False,
             export_metrics=False):

    kwargs = {}
    # if fs_store:
    #     from eve_fsmediastorage import FileSystemMediaStorage
    #     kwargs["media"] = FileSystemMediaStorage

    if app is None:
        from eve import Eve
        app = Eve(settings=settings, auth=auth, **kwargs)

    if swagger:
        # from eve_swagger import swagger as swagger_blueprint
        from eve_swagger import get_swagger_blueprint
        
        swagger_blueprint = get_swagger_blueprint()
        app.register_blueprint(swagger_blueprint, url_prefix=f'/{app.config["API_VERSION"]}')
        app.config['SWAGGER_INFO'] = {
            'title': 'XENON PMT API',
            'version': '1.0',
            'description': 'API for the XENON PMT database',
            'termsOfService': 'https://opensource.org/ToS',
            'contact': {
                'name': 'Yossi Mosbacher',
                'url': 'https://pmts.xenonnt.org',
                "email": "joe.mosbacher@gmail.com"
            },

            'license': {
                'name': 'BSD',
                'url': 'https://github.com/nicolaiarocci/eve-swagger/blob/master/LICENSE',
            
            },
            'schemes': ['http', 'https'],

        }


    if export_metrics:
        from prometheus_flask_exporter import PrometheusMetrics
        PrometheusMetrics(app)

    return app



from eve.auth import BasicAuth, TokenAuth
from flask import current_app
from xepmts.api.server.v1.secrets import ROOT_TOKEN, GLOBAL_READ_TOKEN
import os
from functools import lru_cache
import requests
import logging
import sys

logger = logging.getLogger(__name__)

@lru_cache(maxsize=100)
def get_roles(token):
    roles = []
    accounts = current_app.data.driver.db['accounts']
    account = accounts.find_one({'token': token})
    if account:
        roles = account.get("roles", [])
    else:
        try:
            resp = requests.get(f'https://{os.getenv("DOMAIN", "localhost:8000")}/db_api/roles/{token}', timeout=5)
            if resp.ok:
                data = resp.json()
                roles = data["roles"]
                if roles:
                    accounts.insert_one({
                        "username": data["user"]+"_"+data["token"],
                        "token": data["token"],
                        "roles": roles,
                    })
                
            # print(f"got roles from website: {roles}", file=sys.stderr)
        except Exception as e:
            print(e)

            # print(f"got roles from mongo account: {roles}", file=sys.stderr)
    # if not roles:
        # print("Found no matching roles")
    return set(roles)


class XenonBasicAuth(BasicAuth):
    def check_auth(self, username, password, allowed_roles, resource,
            method):
        return username == os.getenv("API_USER", "xenon") and password == os.getenv("API_PASS", "secretxenon")

class XenonTokenAuth(TokenAuth):
    # def __init__(self, ):
        # from redis import StrictRedis, ConnectionPool
        # self.redis = StrictRedis()
        # self.redis.connection_pool = ConnectionPool.from_url(os.environ.get(
        #     'REDIS_URL',
        #     'redis://localhost:6379'))
        
    def check_auth(self, token, allowed_roles, resource, method):
        if not token:
            return False
        if ROOT_TOKEN and token == ROOT_TOKEN:
            return True
        if all([GLOBAL_READ_TOKEN, token == GLOBAL_READ_TOKEN, method in ["GET", "HEAD"]]):
            return True
        # print(f"Checking roles for token {token}", file=sys.stderr)
        roles = get_roles(token)
        return bool(roles.intersection(allowed_roles))



import param
import panel as pn
import holoviews as hv
import numpy as np
import base64
import pandas as pd
from pandas.api.types import is_numeric_dtype

from xepmts.web_client.src.shared.templates import ListTemplate
from xepmts.web_client.src.shared import config, get_db
from xepmts.web_client.src.shared._menu import session_menu
from xepmts.web_client.src.shared._session import login_required
from xepmts.web_client.src.shared.models import BaseSection, DataLoader, CenterColumn



class TrendViewer(BaseSection):
    xaxis = param.Selector(["date"], default="date")
    yaxis = param.Selector(["gain"], default="gain")
    overlay = param.Selector(["pmt_index"], default="pmt_index")
    groupby = param.Selector(["sector"], default="sector")
    columns = param.Integer(1)
    alpha = param.Number(0.9)
    colormap = param.Selector(hv.plotting.list_cmaps(), default="Plasma")
    extra_columns = param.List(["gain_err"])
    loading = param.Boolean(False)
    
    _plots = param.Parameter()
    build_plots_button = param.Action(lambda self: self._update_view, label="Redraw plots")

    def _init_view(self):
        self._view = pn.Column(self.plots)

    def _update_view(self, *events):
        if self.data is None:
            return
        self.loading = True
        try:
            self._plots = self.make_plots()
        finally:
            self.loading = False

    @param.depends("_plots")
    def plots(self):
        if self._plots is None:
            button = pn.widgets.Button(name="Build plots",
                                       sizing_mode="stretch_width",
                                       height=150,
                                       button_type="primary")
            button.on_click(lambda event: self.make_plots())
            return pn.Column(button)
        return self._plots
    
    def make_holomap(self):
        curve_dict = {}
        for (overlay, groupby), pdf in self.data.groupby([self.overlay, self.groupby]):
            curve_dict[(overlay, groupby)] = hv.Curve(pdf.sort_values(self.xaxis), kdims=[self.xaxis], 
                                                      vdims=[self.yaxis], label=str(overlay), 
                                                      group=str(groupby)).opts(responsive=True,
                                                                               interpolation="steps-mid")
        holomap = hv.HoloMap(curve_dict, kdims=[self.overlay, self.groupby])
        return holomap
    
    def make_plots(self):
        holomap = self.make_holomap()
        overlay = holomap.overlay(self.overlay).opts(legend_position='top', legend_cols=True, legend_offset=(0, -10))
        plots = overlay.layout(self.groupby).cols(self.columns).opts(
            hv.opts.Curve(logy=False, height=300, show_title=False, interpolation="steps-mid",  ),    
        )
        return plots
 
    def settings(self):
        parameters = ["xaxis", "yaxis", "overlay", "groupby", "build_plots_button"]
        widgets = {}
        return pn.Param(self.param, 
                       parameters=parameters,
                       widgets=widgets
                       )


@login_required(pn.state)                    
def view(db):
    # db = pn.state.as_cached(pn.state.curdoc.session_context.id, get_db)
    menu = session_menu(pn.state.curdoc.session_context.id)
    template = ListTemplate(
        title="Heatmap viewer",
        sidebar_footer=menu,
    )
    pn.state.curdoc.theme = template.theme.bokeh_theme

    plotter = TrendViewer()
    loader = DataLoader(client=db, data_plotter=plotter, _logged_in=db.logged_in)

    hv.renderer('bokeh').theme = template.theme.bokeh_theme
    template.main[:] = [pn.panel(c) for c in loader.cards()]
    template.sidebar[:] = [pn.panel(loader.settings)]

    return template


if __name__.startswith("bokeh"):
    view().servable()



"""Top-level package for xepmts."""

__author__ = """Yossi Mosbacher"""
__email__ = 'joe.mosbacher@gmail.com'
__version__ = '0.4.17'

# import eve_panel
from xepmts import api
from xepmts.api.server.v1.app import list_roles
from xepmts.api.server.v2.app import list_roles as list_v2_roles

from xepmts.api.client import default_client, get_client
from . import streams

def settings(**kwargs):
    from eve_panel import settings as panel_settings
    if not kwargs:
        return dir(panel_settings)
    else:
        for k,v in kwargs.items():
            setattr(panel_settings, k, v)


def extension():
    import eve_panel
    eve_panel.extension()

notebook = extension


import xepmts

def get_db():
    xepmts_client = xepmts.get_client("v2")
    xepmts_client.select_server("xenonnt.org_v2")
    # xepmts_client.set_server_url(gae2_v2="https://api2-dot-xenon-pmts.uc.r.appspot.com/v2")
    # xepmts_client.select_server("gae2_v2")
    return xepmts_client




# pylint: disable=redefined-outer-name,protected-access,missing-function-docstring
"""In this module we provide functionality to serve the Site

We provide some basic settings for `panel.serve` below. They might need tweeking to support your
use case.

For more info refer to the [README](README.md) and the section on Deployment.
"""
import os
import platform



from xepmts.web_client.src.shared import config, modifications
modifications.apply()

import holoviews as hv
hv.extension("bokeh")

import panel as pn
import xepmts

xepmts.extension()

def serve(**kwargs):
    """Serves the site

    Use the configuration files together with the
    [BOKEH environment variables]\
    (https://docs.bokeh.org/en/latest/docs/reference/settings.html#bokeh-settings) to configure
    your site and how its served.
    """
    address = os.getenv("BOKEH_ADDRESS", "localhost")
    if platform.system() == "Windows":
        kwargs.pop("num_procs")
    
    pn.serve(
            config.routes,
            port=kwargs.get("port", 5006),
            dev=kwargs.get("debug", False),
            title=config.site_name,
            address=kwargs.get("address", address),
            num_procs=kwargs.get("nproc", 1),
            static_dirs=config.static_dirs,
            allow_websocket_origin=kwargs.get("allow_websocket_origin", ["localhost"]),

        )


if __name__ == "__main__":
    serve()




import pathlib
import os
from xepmts.api.server.v1.utils import read_endpoint_files, resources_from_templates
from xepmts import api

SOURCE_DIR = pathlib.Path(__file__).parent.parent.parent
DEFAULT_TEMPLATE_DIR = SOURCE_DIR / "endpoints"

def get_domain():
    ENDPOINT_TEMPLATE_DIR = os.getenv("XEPMTS_ENDPOINT_TEMPLATE_DIR", DEFAULT_TEMPLATE_DIR.absolute())
    # print(ENDPOINT_TEMPLATE_DIR)
    ENDPOINT_DIR = os.getenv("XEPMTS_ENDPOINT_DIR", "")
    if not ENDPOINT_DIR:
        templates = read_endpoint_files(ENDPOINT_TEMPLATE_DIR)
        DOMAIN = resources_from_templates(templates)
    else:
        DOMAIN = read_endpoint_files(ENDPOINT_DIR)
    return DOMAIN


"""This module provides various modifies the Panel `get_server` function to enable the site to
serve .py and .ipynb files."""
# pylint: skip-file
from __future__ import absolute_import, division, unicode_literals

import pathlib
import signal
import sys
import uuid
from functools import partial

from bokeh.command.util import build_single_handler_application
from bokeh.server.server import Server
from panel.io.server import INDEX_HTML, ProxyFallbackHandler, _eval_panel, get_static_routes
from panel.io.state import state
from tornado.ioloop import IOLoop
from tornado.wsgi import WSGIContainer
from xepmts.web_client import __file__ as src_file

ROOT_PATH = pathlib.Path(src_file).parent


def get_server(
    panel,
    port=0,
    address=None,
    websocket_origin=None,
    loop=None,
    show=False,
    start=False,
    title=None,
    verbose=False,
    location=True,
    static_dirs={},
    oauth_provider=None,
    oauth_key=None,
    oauth_secret=None,
    oauth_extra_params={},
    cookie_secret=None,
    oauth_encryption_key=None,
    **kwargs,
):
    """
    Returns a Server instance with this panel attached as the root
    app.

    Arguments
    ---------
    panel: Viewable, function or {str: Viewable}
      A Panel object, a function returning a Panel object or a
      dictionary mapping from the URL slug to either.
    port: int (optional, default=0)
      Allows specifying a specific port
    address : str
      The address the server should listen on for HTTP requests.
    websocket_origin: str or list(str) (optional)
      A list of hosts that can connect to the websocket.

      This is typically required when embedding a server app in
      an external web site.

      If None, "localhost" is used.
    loop : tornado.ioloop.IOLoop (optional, default=IOLoop.current())
      The tornado IOLoop to run the Server on.
    show : boolean (optional, default=False)
      Whether to open the server in a new browser tab on start.
    start : boolean(optional, default=False)
      Whether to start the Server.
    title : str or {str: str} (optional, default=None)
      An HTML title for the application or a dictionary mapping
      from the URL slug to a customized title.
    verbose: boolean (optional, default=False)
      Whether to report the address and port.
    location : boolean or panel.io.location.Location
      Whether to create a Location component to observe and
      set the URL location.
    static_dirs: dict (optional, default={})
      A dictionary of routes and local paths to serve as static file
      directories on those routes.
    oauth_provider: str
      One of the available OAuth providers
    oauth_key: str (optional, default=None)
      The public OAuth identifier
    oauth_secret: str (optional, default=None)
      The client secret for the OAuth provider
    oauth_extra_params: dict (optional, default={})
      Additional information for the OAuth provider
    cookie_secret: str (optional, default=None)
      A random secret string to sign cookies (required for OAuth)
    oauth_encryption_key: str (optional, default=False)
      A random encryption key used for encrypting OAuth user
      information and access tokens.
    kwargs: dict
      Additional keyword arguments to pass to Server instance.

    Returns
    -------
    server : bokeh.server.server.Server
      Bokeh Server instance running this panel
    """
    server_id = kwargs.pop("server_id", uuid.uuid4().hex)
    kwargs["extra_patterns"] = extra_patterns = kwargs.get("extra_patterns", [])
    if isinstance(panel, dict):
        apps = {}
        for slug, app in panel.items():
            if isinstance(title, dict):
                try:
                    title_ = title[slug]
                except KeyError:
                    raise KeyError(
                        "Keys of the title dictionnary and of the apps "
                        f"dictionary must match. No {slug} key found in the "
                        "title dictionnary."
                    )
            else:
                title_ = title
            slug = slug if slug.startswith("/") else "/" + slug
            if "flask" in sys.modules:
                from flask import Flask

                if isinstance(app, Flask):
                    wsgi = WSGIContainer(app)
                    if slug == "/":
                        raise ValueError("Flask apps must be served on a subpath.")
                    if not slug.endswith("/"):
                        slug += "/"
                    extra_patterns.append(
                        ("^" + slug + ".*", ProxyFallbackHandler, dict(fallback=wsgi, proxy=slug))
                    )
                    continue
                  
            if isinstance(app, str):
                app = pathlib.Path(app)
            if not app.is_absolute():
              app = ROOT_PATH / app

            app = str(app)
            if isinstance(app, str) and (app.endswith(".py") or app.endswith(".ipynb")):
                apps[slug] = build_single_handler_application(app)
            else:
                apps[slug] = partial(_eval_panel, app, server_id, title_, location)
    else:
        apps = {"/": partial(_eval_panel, panel, server_id, title, location)}

    extra_patterns += get_static_routes(static_dirs)

    opts = dict(kwargs)
    if loop:
        loop.make_current()
        opts["io_loop"] = loop
    elif opts.get("num_procs", 1) == 1:
        opts["io_loop"] = IOLoop.current()

    if "index" not in opts:
        opts["index"] = INDEX_HTML

    if address is not None:
        opts["address"] = address

    if websocket_origin:
        if not isinstance(websocket_origin, list):
            websocket_origin = [websocket_origin]
        opts["allow_websocket_origin"] = websocket_origin

    # Configure OAuth
    from panel.config import config

    if config.oauth_provider:
        from panel.auth import OAuthProvider

        opts["auth_provider"] = OAuthProvider()
    if oauth_provider:
        config.oauth_provider = oauth_provider
    if oauth_key:
        config.oauth_key = oauth_key
    if oauth_extra_params:
        config.oauth_extra_params = oauth_extra_params
    if cookie_secret:
        config.cookie_secret = cookie_secret
    opts["cookie_secret"] = config.cookie_secret

    server = Server(apps, port=port, **opts)
    if verbose:
        address = server.address or "localhost"
        print("Launching server at http://%s:%s" % (address, server.port))

    state._servers[server_id] = (server, panel, [])

    if show:

        def show_callback():
            server.show("/login" if config.oauth_provider else "/")

        server.io_loop.add_callback(show_callback)

    def sig_exit(*args, **kwargs):
        server.io_loop.add_callback_from_signal(do_stop)

    def do_stop(*args, **kwargs):
        server.io_loop.stop()

    try:
        signal.signal(signal.SIGINT, sig_exit)
    except ValueError:
        pass  # Can't use signal on a thread

    if start:
        server.start()
        try:
            server.io_loop.start()
        except RuntimeError:
            pass
    return server



import xepmts
import holoviews as hv
import param
import panel as pn
import pandas as pd
import numpy as np
import httpx
import asyncio
import time
import eve_panel
from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager
from functools import cache
from panel.io.server import unlocked
from tornado.ioloop import IOLoop

pn.config.sizing_mode = "stretch_width"
executor = ThreadPoolExecutor(max_workers=4)


def get_field_limit(resource, field="timestamp", sort=""):
    sort = sort + field
    r = resource.find(projection={field: 1}, sort=sort, max_results=1)
    return pd.to_datetime(r[0][field])

def get_date_range(resource, field="timestamp"):
    return get_field_limit(resource, field=field), get_field_limit(resource, field=field, sort="-")

def CenterColumn(*objs, **kwargs):
    for obj in objs:
        obj.align = "center"
#     kwargs["margin"] = kwargs.get("margin", 1)
    return pn.Column(pn.layout.VSpacer(), *objs, pn.layout.VSpacer(), **kwargs)

def CenterRow(*objs, **kwargs):
    for obj in objs:
        obj.align = "center"
    kwargs["margin"] = kwargs.get("margin", 5)
    return pn.Column(pn.layout.HSpacer(), *objs, pn.layout.HSpacer(), **kwargs)


class GainsViewer(param.Parameterized):
    PMT_PLOT_SETTINGS = ["experiment", "detector",
                         "color_column", "alpha", "ncols", "colormap"]
    client = param.Parameter(default=None)
    experiment = param.Selector()
    detector = param.Selector()
    date_range = param.DateRange()
    
    _pmt_selection_plot = param.Parameter(precedence=-1)
    _selected_pmts_table = param.Parameter(precedence=-1)
    _gains_plot = param.Parameter(precedence=-1)
    
    selected = param.DataFrame(precedence=-1)
    
    ncols = param.Integer(2)
    color_column = param.Selector()
    alpha = param.Number(0.5)
    colormap = param.Selector(hv.plotting.list_cmaps(), default="Set1")
    
    load_pmt_selection = param.Action(lambda self: self._load_pmt_selection_pressed(), label="Select pmts")
    load_gains_map = param.Action(lambda self: self._load_gains_pressed(), label="Load data")
    
    gains = param.DataFrame(precedence=-1)
    
    _loading = param.Boolean(False)
    _npages = param.Integer(100)
    _loaded = param.Integer(0)

    _main = param.Parameter(precedence=-1)
    _selections = param.Parameter(precedence=-1)
    _installs = param.Parameter(precedence=-1)
    _installs_cache = param.Dict({})
    _pages = param.List([])
    _cb = param.Parameter(precedence=-1)
    _done_loading = param.Boolean(False)
    _logged_in = param.Boolean(False)
    
    def main_view(self):
        if self._main is None:
            self._main = pn.Column(
                *self.panels(),
                sizing_mode="stretch_both"
                              )
        self._update_menu()
        return self._main
    
    def panels(self):
        return [
            pn.panel(self.pmt_selection_plot),
            pn.panel(self.selected_pmts_table),
            pn.panel(self.gains_plot),
        ]
    
    def __init__(self, **params):
        super().__init__(**params)
        self._update_menu()

    def _update_login_status(self):
        self._logged_in = self.client.logged_in

    @param.depends("_pmt_selection_plot", "_logged_in")
    def pmt_selection_plot(self):
        if self._pmt_selection_plot:
            content = self._pmt_selection_plot
        elif self.client.logged_in:
        
            button = pn.widgets.Button(name="Load PMT maps",
                                       width=400, height=200, 
                                       max_height=200, sizing_mode="stretch_width")
            def cb(event):
                self._load_pmt_selection_pressed()

            button.on_click(cb)
            content = CenterColumn(button, sizing_mode="stretch_both", )
        else:
            
            content = pn.panel("### Database access needed. \nYou must login before using this page.")
        view = pn.Card(content, header="## PMT Selection", sizing_mode="stretch_both",)
        def collapse_view(event):
            if event.old==False and event.new==True:
                view.collapsed = True
        self.param.watch(collapse_view, ["_loading"])

        def reset_view(*events):
            self._pmt_selection_plot = None
            self.gains = None
        self.param.watch(reset_view, ["experiment", "detector"])

        return view

    @param.depends("_selected_pmts_table")
    def selected_pmts_table(self):
        if self._selected_pmts_table is None:
            content = CenterColumn(sizing_mode="stretch_width")
        else:
            content = self._selected_pmts_table
        view = pn.Card(content, header="## Selected PMT details",
                       collapsed=False, sizing_mode="stretch_both")
        def collapse_view(event):
            if event.old==False and event.new==True:
                view.collapsed = True
        self.param.watch(collapse_view, ["_loading"])
        return view
        
    @param.depends("client")
    def _update_experiment_options(self):
        if self.client is None:
            return
        self.param.experiment.names = {"xenonnt": self.client, "xenon1t": self.client.xenon1t}
        self.param.experiment.objects = [self.client, self.client.xenon1t]
        self.experiment = self.client
        
    @param.depends("experiment", watch=True)
    def _update_detector_options(self):
        if self.experiment is None:
            return
        detectors = {k:v for k,v in self.experiment.sub_resources.items() 
                     if k not in self.param.experiment.names}
        self.param.detector.objects = list(detectors.values())
        self.param.detector.names = detectors
        self.detector = detectors.get("tpc",list(detectors.values())[0])
    
    @param.depends("detector","_logged_in", watch=True)
    def _update_plotting_options(self):
        if self.detector is None:
            return
        try:    
            self.param.color_column.objects = list(
                self.detector.installs.find_df(max_results=1).dropna(axis=1, how="all").columns)
        except:
            pass
        self.color_column = "sector"
        try:
            drange = get_date_range(self.detector.gains)
        except:
            drange = pd.to_datetime("21/3/2019"), pd.to_datetime(time.time()*1e9)
        self.param.date_range.softbounds = drange
        self.date_range = drange
        

    @param.depends("_logged_in", watch=True)
    def _update_menu(self):
        if self.experiment is None:
            self._update_experiment_options()
        if self.detector is None:
            self._update_detector_options()
        if self.color_column is None or self.date_range is None:
            self._update_plotting_options()
    
    @property
    def installs(self):
        if self.detector is None:
            return hv.Table(pd.DataFrame({
                "pmt_index": [], "array":[], "position_x": [], "position_y": []}))

        key = self.experiment.name + self.detector.name
        if key not in self._installs_cache:
            installs = self.detector.installs.paginate(300).df.dropna(axis=1, how="all")
            self._installs_cache[key] = hv.Table(installs)
        return self._installs_cache[key]
    
    def _load_gains_pressed(self):
        if self._loading:
            return "Already loading"
        if self._cb is not None and self._cb.running:
            self._cb.stop()
            
        self._loading = True
        self.gains = None
        self._pages = []
        try:
            indices = self.selected_pmts()
            gains = self.detector.gains
            date_min, date_max = self.date_range
            if len(indices):
                gains = gains.paginate(500).filter(
                    pmt_index={"$in": indices},
                    timestamp={"$gt": date_min.timestamp()*1e9, "$lt": date_max.timestamp()*1e9},)
            self._npages = len(gains.page_numbers)
            def cb(idx):
                self._loaded += 1
        except Exception as e:
            self._loading = False
            return e
        loop = IOLoop.current()
        for idx in gains.page_numbers:
            future = executor.submit(gains.get_page, idx)
            loop.add_future(future, self._update_gains)
        self._cb = pn.state.add_periodic_callback(self._check_gains_loaded)

   
    def _update_gains(self, future):
        page = None
        try:
            page = future.result()
            if not self._loading:
                return
        except:
            pass
        finally:
            with unlocked():
                if page:
                    self._pages.append(page)
                self._loaded += 1
            

    def _check_gains_loaded(self):
        if not self._loading:
            self._pages = []
            if self._cb is not None and self._cb.running:
                self._cb.stop()
                return

        if len(self._pages)<self._npages:
            return

        with unlocked():
            try:
                df = pd.concat([page.to_dataframe() for page in self._pages])
                df["pmt_index"] = df["pmt_index"].astype(str)
                df = df.sort_values(["pmt_index", "timestamp"])
                df["date"] = pd.to_datetime(df["timestamp"]).dt.strftime('%Y-%m-%d')
                df = df.replace(-np.inf, np.nan)
                df = df.replace(np.inf, np.nan)
                self.gains = df
            except:
                pass
            finally:
                self._loading = False
            
    def _load_pmt_selection_pressed(self):
        spinner = pn.indicators.LoadingSpinner(value=True, width=150, height=150, align='center')
        self._pmt_selection_plot = CenterColumn(spinner, sizing_mode="stretch_width", )
        pn.state.add_periodic_callback(self.load_pmt_arrays, count=1)
        
    @param.depends("ncols", "color_column", "alpha", "colormap",)
    def pmt_arrays(self):
        installs = self.installs
        cols = [dim.name for dim in installs.dimensions()]
   
        arrays = np.unique(installs["array"])
        extra_cols = [col for col in cols if col not in ["position_x", "position_y"]]
        pmt_plot = hv.Points(installs, kdims=["position_x", "position_y"], vdims=extra_cols).opts(
                    alpha=self.alpha, color=self.color_column, size=15, 
                    nonselection_alpha=0.1, cmap=self.colormap, colorbar=True,
        )
        array_plots = {array: pmt_plot.select(array=array).opts(default_tools=["tap", "hover", "lasso_select"], 
                                                                aspect=1.1, height=300, responsive=True) for array in arrays}
        self._selections = [hv.streams.Selection1D(source=array_plots[array], throttle_timeout=1000, throttle_scheme="throttle").rename(index=f"{array}") for array in arrays ]
        
        def make_table(**kwargs):
            dfs = []
            for array, idx in kwargs.items():
                if idx:
                    dfs.append(self.installs.select(array=array).iloc[idx].dframe())
                else:
                    dfs.append(self.installs.select(array=array).dframe())
            if not len(dfs):
                return self.installs
            df = pd.concat(dfs).reset_index()
            table = hv.Table(df).opts(
                hv.opts.Table(width=900, fit_columns=True)
            )
            return table

        self._selected_pmts_table = CenterColumn(hv.DynamicMap(make_table, streams=self._selections),
                                                    sizing_mode="stretch_width") 
        
        
        return hv.NdLayout(array_plots, kdims="array").cols(self.ncols)
   
    def load_pmt_arrays(self):
        button = pn.widgets.Button(name="Click to load gains for selection",
                                       width=400, height=70,
                                       max_height=100, sizing_mode="stretch_width")
        def load_gains(event):
            self._load_gains_pressed()

        button.on_click(load_gains)
        plot_wrapper = pn.Column(self.pmt_arrays,
                                 sizing_mode="stretch_both", height=370, margin=10)
        self._pmt_selection_plot = pn.Column(plot_wrapper, 
                                            pn.Param(self.param.date_range),
                                            button,
                                            height=500,
#                                                 max_height=500,
                                            sizing_mode="stretch_both")
    
    def selected_pmts(self):
        if not self._selections:
            return []
        installs = self.installs
        pmt_indices = []
        for s in self._selections:
            array = s._rename["index"]
            if s.index:
                table = installs.select(array=array).iloc[s.index]
            else:
                table = installs.select(array=array)
            pmt_indices.extend([int(x) for x in table.dimension_values("pmt_index")])
        return pmt_indices
    

    @param.depends("_loaded", "_loading")
    def progress(self):
        bar = pn.indicators.Progress(value=self._loaded, active=self._loading, max=self._npages,
                                    sizing_mode="stretch_width", margin=10, min_width=300)
        perc = int(100*self._loaded/self._npages)
        text = pn.widgets.StaticText(value=f"Loading requested data. {perc}% [{self._loaded}/{self._npages} pages] done.", 
                                    align="center")
        return CenterColumn(text, bar, sizing_mode="stretch_width",)

    @param.depends("gains", "_loading")
    def gains_plot(self):
        if self._loading:
            cancel = pn.widgets.Button(name="Cancel", sizing_mode="stretch_width")
            def cancel_pressed(event):
                self._loading = False
            cancel.on_click(cancel_pressed)
            
            progress = pn.panel(self.progress)
            content = pn.Column(
                                progress,
                                cancel,
                                sizing_mode="stretch_width")

        elif self.gains is None or not len(self.gains):
            content = CenterColumn(pn.panel("Gains visualizations will be shown here after being loaded."),
                                   sizing_mode="stretch_both")
        else:
            aspect = len(self.gains["date"].unique())/len(self.gains["pmt_index"].unique())
            hm = hv.HeatMap(self.gains, kdims=["date", "pmt_index"], vdims=["gain", "gain_err"]).aggregate(function=np.mean)
            hm = hm.opts(tools=["tap", "hover"], nonselection_alpha=0.1, aspect=aspect, responsive=True
                          ,xrotation=60, xaxis="top")
            content = pn.Column(hm, sizing_mode="stretch_both", height=600, width=700)
        return pn.Card(content, header="## Gains plot", sizing_mode="stretch_both")


    def settings_view(self):
        return pn.Column(pn.Param(self.param, expand_button=False,  width=250,
                        parameters=self.PMT_PLOT_SETTINGS), width=270, sizing_mode="fixed")
    


from werkzeug.wsgi import pop_path_info, peek_path_info

class PathDispatcher:

    def __init__(self, base_app, app_dict):
        self.base_app = base_app
        self.apps = app_dict

    def __call__(self, environ, start_response):
        app = self.apps.get(peek_path_info(environ), self.base_app)
        return app(environ, start_response)



"""## The Landing Page of the site"""
import pathlib

import panel as pn
from panel.pane import Markdown
from awesome_panel_extensions.frameworks.fast.fast_menu import to_menu

from xepmts.web_client.src.shared.templates import ListTemplate
from xepmts.web_client.src.shared import config, get_db
from xepmts.web_client.src.shared._menu import session_menu
from xepmts.web_client.src.shared.models import CenterColumn
from copy import deepcopy

HOME_PATH = pathlib.Path(__file__).parent / "home.md"
HOME_CONTENT = HOME_PATH.read_text()
AINT_NOONE_GOT_TIME = str(HOME_PATH.parent / "aint.gif")

WELCOME = """
### Welcome to the XENON PMT archive, home to all things PMT related.
Please send requests for additional content to [Yossi Mosbacher](mailto:joe.mosbacher@gmail.com)



"""
def view():
    """Returns the landing page of the site"""
    pn.config.sizing_mode = "stretch_width"
    db = pn.state.as_cached(pn.state.curdoc.session_context.id, get_db)
    db.session.auth.token_file = ""
    aint_noone = CenterColumn("# Writing a python script just to check the PMT gain evolution?", 
                               pn.pane.GIF(AINT_NOONE_GOT_TIME),
                               "# Thats what the web client is for.")
    sections = [
        pn.pane.Markdown(WELCOME),
        pn.Card(db.session.panel(), header="## Authentication"),
        pn.pane.Markdown(HOME_CONTENT),
        aint_noone,
    ]
    
    menu = session_menu(pn.state.curdoc.session_context.id)
    return ListTemplate(title="Home", main=sections, main_max_width="900px", sidebar_footer=menu)


if __name__.startswith("bokeh"):
    # Run the development server
    # python -m panel serve 'src/apps/home/home.py' --dev --show
    view().servable()
if __name__ == "__main__":
    # Run the server. Useful for integrated debugging in your Editor or IDE.
    # python 'src/apps/home/home.py'
    view().show(port=5007)



"""Provides

- config: An instance of SiteConfig holding the current Site configuration
"""

import pathlib
from typing import Union
import panel as pn
import xepmts

import toml

from xepmts.web_client.src.shared.models import Application, Person, SiteConfig
from xepmts.web_client import __file__ as src_file

ROOT_PATH = pathlib.Path(src_file).parent

def _create_site_config(config_root: Union[str, pathlib.Path]) -> SiteConfig:
    if isinstance(config_root, str):
        config_root = pathlib.Path(config_root)
    persons_toml = config_root / "persons.toml"
    persons = Person.create_from_toml(persons_toml)

    applications = config_root / "applications.toml"
    applications = Application.create_from_toml(applications, persons=persons)
    # print(applications)
    # for app in applications:
    #     app.code_url = ROOT_PATH / app.code_url
    
    site_toml = config_root / "site.toml"
    site = toml.load(site_toml)
    if "static_dirs" in site:
        for k,v in site["static_dirs"].items():
            apath = pathlib.Path(v)
            if not apath.is_absolute():
                apath = ROOT_PATH / apath
            site["static_dirs"][k] = str(apath)
    
    # def save_token(auth):
    #     pn.state.
    
    # xepmts_client.session.auth.post_login_hooks.append()
    return SiteConfig(**site, persons=persons, applications=applications)



import os
import xepmts

from flask import Flask


from threading import Lock
from werkzeug.wsgi import pop_path_info, peek_path_info
from werkzeug.serving import run_simple as _run_simple

from xepmts.api.server.v1.app import make_app as make_v1_app
from xepmts.api.server.v1.auth import XenonTokenAuth
from xepmts.api.server.v2.app import make_app as make_v2_app
from xepmts.api.server.utils import PathDispatcher

def create_app():
    from eve_jwt import JWTAuth
    from flask_swagger_ui import get_swaggerui_blueprint
    from prometheus_flask_exporter import PrometheusMetrics

    v1 = make_v1_app(auth=XenonTokenAuth, swagger=True)
    v2 = make_v2_app(auth=JWTAuth, swagger=True)
    app_versions = {
        "v1": v1, 
        "v2": v2
        }
    app = Flask(__name__)
    # @app.route("/")
    # def hello():
    #     return "You have reached the PMT db."

    app.config['SWAGGER_INFO'] = {
            'title': 'XENON PMT API',
            'version': '1.0',
            'description': 'API for the XENON PMT database',
            'termsOfService': 'https://opensource.org/ToS',
            'contact': {
                'name': 'Yossi Mosbacher',
                'url': 'https://pmts.xenonnt.org',
                "email": "joe.mosbacher@gmail.com"
            },

            'license': {
                'name': 'BSD',
                'url': 'https://github.com/nicolaiarocci/eve-swagger/blob/master/LICENSE',
            
            },
            'schemes': ['http', 'https'],

        }
    config = {
        'app_name': "PMT Database API",
        "urls": [{"name": f"Xenon PMT Database {v.capitalize()}", "url": f"/{v}/api-docs" } for v in app_versions]
    }
    API_URL = '/v2/api-docs'
    SWAGGER_URL = ''
    SWAGGERUI_BLUEPRINT = get_swaggerui_blueprint(
        SWAGGER_URL,
        API_URL,
        config=config,
    )
    app.register_blueprint(SWAGGERUI_BLUEPRINT, url_prefix=SWAGGER_URL)
    # PrometheusMetrics(app)

    application = PathDispatcher(app,
                         app_versions)

    return application

def run_simple(address='0.0.0.0', port=5000, debug=True, reload=True, evalex=True):
    app = create_app()
    _run_simple(address, port, app,
                use_reloader=debug, use_debugger=reload, use_evalex=evalex)

if __name__ == '__main__':
    run_simple()
    




import param
import panel as pn
import holoviews as hv
import numpy as np
import base64
import pandas as pd
from pandas.api.types import is_numeric_dtype

from xepmts.web_client.src.shared.templates import ListTemplate
from xepmts.web_client.src.shared import config, get_db
from xepmts.web_client.src.shared._menu import session_menu
from xepmts.web_client.src.shared._session import login_required
from xepmts.web_client.src.shared.models import BaseSection, DataLoader, CenterColumn


def embed_pdf(pdf):
    if isinstance(pdf, bytes):
        pdf = base64.b64encode(pdf).decode()
    html = f"""
    <object width=100% height=100% data="data:application/pdf;base64,{pdf}" type="application/pdf">
    <p>Looks like your browser doesnt support embedding pdfs. Maybe time to upgrade?</p>
    </object>
    """
    return html

class DatasheetBrowser(param.Parameterized):
    client = param.Parameter()
    experiment = param.Selector()
    detector = param.Selector()
    data = param.DataFrame()
    _pmt_list = param.Parameter()
    _view = param.Parameter()
    _datasheet_view = param.Parameter()
    loading = param.Parameter()
    load_pmts_button = param.Action(lambda self: self.update_pmt_list(), label="Show PMTs")
    show_datasheet_button = param.Action(lambda self: self.show_datasheet(), label="Show datasheet")
    
    def __init__(self, **params):
        super().__init__(**params)
        self.update_menu()
    
    @param.depends("client", watch=True)
    def _update_experiment_options(self):
        if self.client is None:
            return
        experiments = {"xenonnt": self.client,
                    "xenon1t": self.client.xenon1t}
        self.param.experiment.names = experiments
        self.param.experiment.objects = [self.client, self.client.xenon1t]
        if self.experiment not in experiments.values():
            self.experiment = self.client
        
    @param.depends("experiment", watch=True)
    def _update_detector_options(self):
        if self.experiment is None:
            return
        detectors = {k:v for k,v in self.experiment.sub_resources.items() 
                     if k not in self.param.experiment.names}
        self.param.detector.objects = list(detectors.values())
        self.param.detector.names = detectors
        if self.detector not in detectors.values():
            self.detector = detectors.get("tpc", list(detectors.values())[0])
    
    def update_menu(self):
        with param.discard_events(self):
            self._update_experiment_options()
            self._update_detector_options()
            
    def update_pmt_list(self):
        if self.detector is None:
            return
        self.loading = True
        try:
            pmts = self.detector.pmts.project(datasheet=0).to_dataframe()
            installs = self.detector.installs.df
            self.data = pmts.merge(installs)
            self._pmt_list = pn.widgets.DataFrame(self.data, sizing_mode="stretch_width", height=250)
        finally:
            self.loading = False
            
    def download_selected(self):
        if self._pmt_list is None or not self._pmt_list.selection:
            return b""
        data = self.data.reset_index()
        sn = data["serial_number"].iloc[self._pmt_list.selection[0]]
        query = {"serial_number": sn}
        return self.detector.pmts.find_one_item(query=query).datasheet
    
    def show_datasheet(self):
        self.loading = True
        try:
            pdf = self.download_selected()
            if pdf:
                self._datasheet_view = pn.pane.HTML(embed_pdf(pdf),
                                                    sizing_mode="stretch_both")
        finally:
            self.loading = False
    
    @param.depends("_pmt_list", "loading")
    def pmt_list(self):
        if self.loading:
            return CenterColumn(pn.indicators.LoadingSpinner(value=True),
                             sizing_mode="stretch_both" )
        if self._pmt_list is None:
            button = pn.widgets.Button(name="Load PMT list", button_type="primary",
                                 sizing_mode="stretch_width", height=200)
            button.on_click(lambda event: self.update_pmt_list())
            return pn.Column(button, sizing_mode="stretch_width")
        else:
            load_button = pn.widgets.Button(name="Load Selected Datasheet", button_type="primary",
                                 sizing_mode="stretch_width", height=50)
            load_button.on_click(lambda event: self.show_datasheet())
            return pn.Column(self._pmt_list,
                             load_button,
                             sizing_mode="stretch_both")
        
    @param.depends("_datasheet_view",)
    def datasheet_view(self):
        if self._datasheet_view is None:
            return CenterColumn(pn.panel("# No datasheet loaded yet.", sizing_mode="stretch_both"), sizing_mode="stretch_both")
        return pn.Column(self._datasheet_view,
                        sizing_mode="stretch_width",
                        height=900)
    
    @param.depends("data",)
    def view(self):
        return pn.Column(self.pmt_list,
                        self.datasheet_view,
                        sizing_mode="stretch_both")
    @property
    def sections(self):
        return {
            "Datasheet List": self.pmt_list,
            "Datasheet viewer": self.datasheet_view,
        }
    
    def settings(self):
        
        return pn.Param(self.param,
                       parameters=["experiment", "detector", "load_pmts_button"],
                       
                       expand_button=False)

@login_required(pn.state)                    
def view(db):
    # db = pn.state.as_cached(pn.state.curdoc.session_context.id, get_db)
    menu = session_menu(pn.state.curdoc.session_context.id)
    template = ListTemplate(
        title="PMT Datasheets",
        sidebar_footer=menu,
    )
    pn.state.curdoc.theme = template.theme.bokeh_theme

    browser = DatasheetBrowser(client=db)

    hv.renderer('bokeh').theme = template.theme.bokeh_theme
    template.main[:] = [pn.Card(v, header="## "+k, sizing_mode="stretch_both") for k,v in browser.sections.items()]
    template.sidebar[:] = [pn.panel(browser.settings)]
    return template


if __name__.startswith("bokeh"):
    view().servable()




import param
import panel as pn
import holoviews as hv
import numpy as np
import base64
import pandas as pd
from pandas.api.types import is_numeric_dtype
import streamz
from streamz.dataframe import PeriodicDataFrame
import httpx
import hvplot
import os

from xepmts.web_client.src.shared.templates import ListTemplate
from xepmts.web_client.src.shared import config, get_db
from xepmts.web_client.src.shared._menu import session_menu
from xepmts.web_client.src.shared._session import login_required
from xepmts.web_client.src.shared.models import BaseSection, FilterTool, CenterColumn
from xepmts.streams import get_live_rate_viewer

class RatesViewer(param.Parameterized):
    client = param.Parameter()
    api_user = param.String()
    api_key = param.String()
    stream_viewer = param.Parameter()
    detectors = param.ListSelector(default=["tpc", "nveto", "muveto"],
     objects=["tpc", "nveto", "muveto"])
    get_viewer_button = param.Action(lambda self: self.get_viewer(), label="Generate viewer")
    loading = param.Boolean(False)

    def get_viewer(self):
        self.loading = True
        try:
            self.stream_viewer = get_live_rate_viewer(db=self.client,
                                                  api_user=self.api_user, 
                                                  api_key=self.api_key,
                                                  detectors=self.detectors)
        finally:
            self.loading = False

    def credentials_view(self):
        return pn.Column(self.param.api_user,
                         self.param.api_key)

    @param.depends("stream_viewer", "loading")
    def view(self):
        if self.loading:
            return CenterColumn(pn.indicators.LoadingSpinner(value=True))

        if self.stream_viewer is None:
            return self.settings()
        else:
            return self.stream_viewer.view()

    def settings(self):
        return pn.Column(self.credentials_view(),
                    self.param.detectors,
                    self.param.get_viewer_button,
                            )

    @property
    def sections(self):
        return {"Live DAQ streams": self.view}
    

@login_required(pn.state)          
def view(db):
    # db = pn.state.as_cached(pn.state.curdoc.session_context.id, get_db)
    menu = session_menu(pn.state.curdoc.session_context.id)
    template = ListTemplate(
        title="Live data rates",
        sidebar_footer=menu,
    )
    pn.state.curdoc.theme = template.theme.bokeh_theme
    api_user = os.getenv("DAQ_API_USER", "")
    api_key = os.getenv("DAQ_API_KEY", "")

    browser = RatesViewer(client=db, api_user=api_user, api_key=api_key)

    hv.renderer('bokeh').theme = template.theme.bokeh_theme
    template.main[:] = [CenterColumn("## "+k, v ,sizing_mode="stretch_both") for k,v in browser.sections.items()]
    template.sidebar[:] = [pn.panel(browser.settings)]
    return template


if __name__.startswith("bokeh"):
    view().servable()



import pathlib
import sqlite3 as sql
from os.path import dirname, join

import numpy as np
import pandas as pd
import panel as pn
import holoviews as hv
import param
from datetime import datetime
import time

from awesome_panel_extensions.widgets.perspective_viewer import PerspectiveViewer

from xepmts.web_client.src.shared.templates import ListTemplate
from xepmts.web_client.src.shared import config, get_db
from xepmts.web_client.src.shared._menu import session_menu
from xepmts.web_client.src.shared.models import BaseSection, DataLoader
from xepmts.web_client.src.shared._session import login_required              

@login_required(pn.state)                                 
def view(db):
    menu = session_menu(pn.state.curdoc.session_context.id)
    template = ListTemplate(
        title="Heatmap viewer",
        sidebar_footer=menu,
    )
    pn.state.curdoc.theme = template.theme.bokeh_theme

    hv.renderer('bokeh').theme = template.theme.bokeh_theme
    
    template.main[:] = [db.panel()]
    # template.sidebar[:] = [pn.panel(), pn.layout.Divider()]

    return template

if __name__.startswith("bokeh"):
    view().servable()



"""Provides customized and branded versions of the templates to be used in the site

Currently

- ListTemplate similar to VanillaTemplate shipped with Panel
- GridTemplate similar to ReactTemplate shipped with Panel
- GalleryTemplate a template for creating a Gallery.
"""
# This will make it much easier to customize your templates when you start wanting to brand your
# site.
import param
from awesome_panel_extensions.frameworks.fast.templates.fast_gallery_template import (
    FastGalleryTemplate as _FastGalleryTemplate,
)
from awesome_panel_extensions.frameworks.fast.templates.fast_grid_template import (
    FastGridTemplate as _FastGridTemplate,
)
from awesome_panel_extensions.frameworks.fast.templates.fast_list_template import (
    FastListTemplate as _FastListTemplate,
)
# from panel.template import FastGridTemplate as _FastGridTemplate
# from panel.template import FastListTemplate as _FastListTemplate

from xepmts.web_client.src.shared import config
from xepmts.web_client.src.shared._menu import MENU


class ListTemplate(_FastListTemplate):
    __doc__ = _FastListTemplate.__doc__

    site = param.String(config.site_name)
    sidebar_footer = param.String(MENU)


class GridTemplate(_FastGridTemplate):
    __doc__ = _FastGridTemplate.__doc__

    site = param.String(config.site_name)
    sidebar_footer = param.String(MENU)


# class GalleryTemplate(_FastGalleryTemplate):
#     __doc__ = _FastGalleryTemplate.__doc__

#     site = param.String(config.site_name)
#     title = param.String("Gallery")
#     description = param.String(config.gallery_description)



"""Provides the following models used to create the site

- Application
- Person
- Resource
- SiteConfig

"""

from copy import deepcopy

import param
import panel as pn
import holoviews as hv
import pandas as pd
import numpy as np
import logging
import time

from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager
from panel.io.server import unlocked
from tornado.ioloop import IOLoop

from awesome_panel_extensions.site.application import Application
from awesome_panel_extensions.site.author import Author as Person
from awesome_panel_extensions.site.resource import Resource
from awesome_panel_extensions.site.site_config import SiteConfig as _ApSiteConfig
from awesome_panel_extensions.io.loading import start_loading_spinner, stop_loading_spinner

from eve_panel.utils import to_json_compliant

logger = logging.getLogger(__name__)
executor = ThreadPoolExecutor(max_workers=8)

class SiteConfig(_ApSiteConfig):
    pass
    
def CenterColumn(*objs, **kwargs):
    for obj in objs:
        if hasattr(obj, "align"):
            obj.align = "center"
    kwargs["margin"] = kwargs.get("margin", 5)
    return pn.Column(pn.layout.VSpacer(), *objs, pn.layout.VSpacer(), **kwargs)

def CenterRow(*objs, **kwargs):
    for obj in objs:
        if hasattr(obj, "align"):
            obj.align = "center"
    kwargs["margin"] = kwargs.get("margin", 5)
    return pn.Column(pn.layout.HSpacer(), *objs, pn.layout.HSpacer(), **kwargs)

def get_field_limit(resource, field="timestamp", sort=""):
    sort = sort + field
    r = resource.find(projection={field: 1}, sort=sort, max_results=1)
    return pd.to_datetime(r[0][field])

def get_date_range(resource, field="timestamp"):
    return get_field_limit(resource, field=field), get_field_limit(resource, field=field, sort="-")

class BaseSection(param.Parameterized):
    """Abstract Class for Section of App"""

    data = param.DataFrame(allow_None=True)
    loading = param.Boolean(default=False)
    _view = param.ClassSelector(class_=pn.Column)
    
    @param.depends("loading", "_view")
    def view(self):
        if self.loading or self._view is None:
            return CenterColumn(pn.indicators.LoadingSpinner(value=self.loading),
                                sizing_mode="stretch_both")
        else:
            return self._view

    @property
    def sections(self):
        return {"View": self.view}

    def cards(self):
        cards = [pn.Card(section, header="### "+name, sizing_mode="stretch_both")
                 for name,section in self.sections.items()]
        return cards

    def __init__(self, **params):
        super().__init__(**params)
        self._init_view()
        self._update_view()
        self.param.watch(self._update_view, "data")

    def _init_view(self):
        raise NotImplementedError()

    def _update_view(self, *events):
        raise NotImplementedError()

        
DEFAULT_DRANGE = pd.to_datetime("21/3/2019"), pd.to_datetime(time.time()*1e9)

class FilterTool(BaseSection):
    client = param.Parameter(precedence=-1)
    experiment = param.Selector(precedence=0)
    detector = param.Selector(precedence=1)
    reload_info_action = param.Action(lambda self: self.load_pmt_data(), label="Reload PMT info", precedence=2)
    resource = param.Selector(precedence=3)

    date_range = param.DateRange(default=DEFAULT_DRANGE, softbounds=DEFAULT_DRANGE, precedence=1)
    filters = param.Dict(default={}, precedence=2)
    projection = param.List([], precedence=3)
    color_selection = param.List([], precedence=4)

    xaxis = param.Selector(["position_x"], default="position_x", precedence=4)
    yaxis = param.Selector(["position_y"], default="position_y", precedence=5)
    time_field = param.Selector(["timestamp", "date", "datetime"], default="timestamp", precedence=6)
    color = param.Selector(["sector"], default="sector", precedence=7)
    groupby = param.Selector(["array"], default="array", precedence=6)
    ncols = param.Integer(2, precedence=8)
    size = param.Integer(12, precedence=8)
    alpha = param.Number(0.5, precedence=8)
    colormap = param.Selector(hv.plotting.list_cmaps(), default="Set1", precedence=7)
    
    _selections = param.Parameter(precedence=-1)

    _group_plots = param.Parameter(precedence=-1)
    _details_table = param.Parameter(precedence=-1)

    _installs_cache = param.Dict({}, precedence=-1)

    update_plot_action = param.Action(lambda self: self.update_plot(), label="Update PMT Plot", precedence=9)
    
    
    
    def _init_view(self):
        self._view = pn.Column("PMT data not loaded yet.", sizing_mode="stretch_both")
        self._update_menu()
        
    def _update_view(self, *events):
        if self.data is None:
            pass
        else:
            cols = list(self.data.columns)
            self.param.xaxis.objects = cols
            self.param.yaxis.objects = cols
            self.param.color.objects = cols
            self.param.groupby.objects = cols
            self.update_plot()
            
    @param.depends("loading")
    def view(self):
        if self.loading:
            return CenterColumn(pn.indicators.LoadingSpinner(value=self.loading),
                                sizing_mode="stretch_width")
        if self.data is None:
            button = pn.widgets.Button(name="Select PMTs", button_type="primary",
                                    sizing_mode="stretch_width", height=150)
            def cb(event):
                self.load_pmt_data()
            button.on_click(cb)
            return CenterColumn(button, sizing_mode="stretch_width")
        else:
            return pn.panel(self._group_plots)

    def _update_experiment_options(self):
        if self.client is None:
            return
        
        experiments = {"xenonnt": self.client, "xenon1t": self.client.xenon1t}
        self.param.experiment.names = experiments
        self.param.experiment.objects = [self.client, self.client.xenon1t]
        if self.experiment not in experiments.values():
            self.experiment = experiments.get("xenonnt", list(experiments.values())[0])
        
    @param.depends("experiment", watch=True)
    def _update_detector_options(self):
        if self.experiment is None:
            return
        
        detectors = {k:v for k,v in self.experiment.sub_resources.items() 
                     if k not in self.param.experiment.names}
        self.param.detector.names = detectors
        self.param.detector.objects = list(detectors.values())
        if self.detector not in detectors.values():
            self.detector = detectors.get("tpc", list(detectors.values())[0])
        
    @param.depends("detector", watch=True)
    def _update_resource_options(self):
        if self.detector is None:
            return
        
        resources = self.detector.resources
        self.param.resource.names = resources
        self.param.resource.objects = list(resources.values())
        if self.resource not in resources.values():
            self.resource = resources.get("gains",list(resources.values())[0])
    
    @param.depends("resource", watch=True)
    def _update_date_range_options(self):
        try:
            drange = get_date_range(self.resource, field=self.time_field)
        except:
            drange = pd.to_datetime("21/3/2019"), pd.to_datetime(time.time()*1e9)
        self.param.date_range.softbounds = drange
        self.date_range = drange
       
    def _update_plotting_options(self):
        if self.detector is None:
            return
        cols = list(self.detector.installs.schema)
        self.param.xaxis.objects = cols
        self.param.yaxis.objects = cols
        self.param.color.objects = cols
        self.param.groupby.objects = cols
        
    def _update_menu(self):
        with param.batch_watch(self):
            self._update_experiment_options()
            self._update_detector_options()
            self._update_resource_options()
#             self._update_date_range_options()
            self._update_plotting_options()
        
    def update_plot(self):
        self.loading = True
        try:
            self._group_plots = self.make_group_plots()
            self._details_table = self.make_details_table()
        finally:
            self.loading = False

    def make_group_plots(self):
        installs = hv.Table(self.data)
        cols = list(self.data.columns)
        groups = list(sorted(self.data[self.groupby].unique(), reverse=True))
        extra_cols = [col for col in cols if col not in [self.xaxis, self.yaxis]]
        
        pmt_plot = hv.Points(installs, kdims=[self.xaxis,self.yaxis], vdims=extra_cols).opts(
                    alpha=self.alpha, color=self.color, size=self.size, responsive=True,
                    nonselection_alpha=0.1, cmap=self.colormap, colorbar=True, default_tools=["tap", "hover", "lasso_select"],
        )
        labels = hv.Labels(installs, kdims=[self.xaxis,self.yaxis], vdims=["pmt_index", self.groupby])

        plots = {group: pmt_plot.select(**{self.groupby: group}) for group in groups}
        self._selections = [hv.streams.Selection1D(source=plots[group],).rename(index=f"{group}")
                            for group in groups ]                        
        for k,v in plots.items():
            plots[k] = v*labels.select(**{self.groupby:k})
        if self.xaxis == "position_x" and self.yaxis=="position_y":
            tpc_wall = hv.Ellipse(0,0, 134, kdims=["position_x", "position_y"]).opts(color="white")
            for k,v in plots.items():
                plots[k] = tpc_wall*v

        layout =  hv.NdLayout(plots, kdims=self.groupby).cols(self.ncols).opts(
            hv.opts.Points(framewise=True, aspect=1.1, min_width=350, responsive=True,
                            default_tools=["tap", "hover", "lasso_select"]),
            hv.opts.Ellipse(default_tools=[]),
            hv.opts.Labels(text_font_size="6px", default_tools=[]),
            )
        
        return layout

    def make_details_table(self):
        installs = hv.Table(self.data)
        def make_table(**kwargs):
            dfs = []
            for group, idx in kwargs.items():
                if idx:
                    dfs.append(installs.select(**{self.groupby: group}).iloc[idx].dframe())
                else:
                    dfs.append(installs.select(**{self.groupby: group}).dframe())
            if not len(dfs):
                return installs
            df = pd.concat(dfs).reset_index()
            table = hv.Table(df).opts(
                hv.opts.Table(width=900, fit_columns=True)
            )
            return table

        return hv.DynamicMap(make_table, streams=self._selections)

    def details_table(self):
        if self._details_table is None:
            return pn.Column()
        else:
            return self._details_table

    def selected_pmts(self):
        if self._selections is None:
            return []
        installs = hv.Table(self.data)
        pmt_indices = []

        for s in self._selections:
            group = s._rename["index"]
            filters = {self.groupby: group}
            if s.index:
                table = installs.select(**filters).iloc[s.index]
            else:
                table = installs.select(**filters)
            if len(self.color_selection):
                table = table.select(**{self.color: self.color_selection})
            pmt_indices.extend([int(x) for x in table.dimension_values("pmt_index")])
        return pmt_indices

    @param.depends("color", "data")
    def extra_filters_section(self):
        if self.data is None:
            options = []
        else:
            options = [to_json_compliant(x) for x in self.data[self.color].unique()][:50]
        name = self.color
        if not name.endswith("s"):
            name += "s"
        multi_select = pn.widgets.MultiChoice(name=name, value=options,
                                            options=options, height=200)

        widgets = {
            "date_range": {"type": pn.widgets.DateRangeSlider,
                           "sizing_mode": "stretch_width",
                           },
            "color_selection": multi_select

        }
        parameters = ["date_range", "color_selection" ,"filters"]
        return pn.Param(self.param,
                        parameters=parameters,
                        widgets=widgets,
                        expand_button=False,
                        show_name=False,
                        sizing_mode="stretch_width")

    
    def settings(self):
        return pn.Column(
                    "### Resource settings",
                    self.resource_settings,
                    pn.layout.Divider(margin=1),
                    "### PMT plot",
                    self.plot_settings,
                    width=270, sizing_mode="fixed")
        
    
    def resource_settings(self):
        parameters = ["experiment", "detector", "resource", "reload_info_action",]
        widgets = {}
        return pn.Param(self.param, 
                        parameters=parameters,
                        widgets=widgets,
                        expand_button=False,
                        width=250, show_name=False,)
        
        
    def plot_settings(self):
        parameters = ["groupby","xaxis","yaxis","ncols","color","alpha",
                        "colormap","size", "update_plot_action"]
        widgets = {}
        return pn.Param(self.param, 
                        parameters=parameters,
                        widgets=widgets,
                        expand_button=False,
                        width=250, show_name=False,)

    def pre_pmt_load(self):
        self.loading = True
       
    def on_pmt_load_success(self):
        self.pmt_data_loaded = True
        
    def post_pmt_load(self):
        self.loading = False
        
    def load_pmt_data(self):
        if self.client is None or not self.client.logged_in:
            return
        if self.detector is None:
            return
        self.pre_pmt_load()
     
        key = self.experiment.name + self.detector.name
        if key in self._installs_cache:
            self.data = self._installs_cache[key]
            self.post_pmt_load()
        else:
            resource = self.detector.installs
            future = executor.submit(resource.to_dataframe)
            def cb(future):
                try:
                    data = future.result()
                    self.data = data
                    self._installs_cache[key] = data
                    self.on_pmt_load_success()
                finally:
                    self.post_pmt_load()
            IOLoop.current().add_future(future, cb)

    def get_filtered_resource(self):
        resource = self.resource
        date_min, date_max = self.date_range
        filters = dict(self.filters)
        if self.time_field in resource.schema:
            filters[self.time_field] = {"$gt": date_min.timestamp()*1e9, "$lt": date_max.timestamp()*1e9}
        indices = self.selected_pmts()
        if len(indices):
            filters["pmt_index"] = {"$in": indices}
        resource = resource.filter(**filters)
        return resource

    @property
    def sections(self):
        self._update_menu()
        return {
            "PMT Selection": self.view,
            "Selection details": self.details_table,
            "Additional filters": self.extra_filters_section,
        }
      
    def panels(self):
        return [pn.panel(s) for s in self.sections.values()]
    
    def cards(self):
        cards = [pn.Card(section, header="### "+name, sizing_mode="stretch_both") for name,section in self.sections.items()]
        cards[1].collapsed = True
        return cards


class DataLoader(param.Parameterized):
    filter_tool = param.ClassSelector(FilterTool)
    data_plotter = param.ClassSelector(BaseSection)
    client = param.Parameter(default=None)
    
    merge_pmt_info = param.Boolean(True)
    _data_plot = param.Parameter(precedence=-1)
    data = param.DataFrame()
    resource_name = param.String()
    partition_size = param.Integer(500)
    
    loading = param.Boolean(False)
    _npages = param.Integer(100)
    _ndocs = param.Integer(100)
    _loaded = param.Integer(0)
    
    _view = param.ClassSelector(pn.Column, precedence=-1)
    _selections = param.Parameter(precedence=-1)
    
    _loaded_data_cache = param.List([], precedence=-1)
    _cb = param.Parameter(precedence=-1)
    _logged_in = param.Boolean(False)

    def __init__(self, **params):
        params["filter_tool"] = params.pop("filter_tool", FilterTool(client=params["client"]))
        super().__init__(**params)

    def view(self):
        if self._view is None:
            self._view = pn.Column(
                *self.cards(),
                sizing_mode="stretch_both"
                              )
        return self._view
    
    @property
    def sections(self):
        sections = {}
        sections.update(self.filter_tool.sections)
        sections["Data loading"] = self.loader_section
        sections.update(self.data_plotter.sections)
        return sections

    @param.depends("loading")
    def loader_section(self):
        if self.loading:
            return self.progress
        else:
            load_data = pn.widgets.Button(name="Load data", 
            height=50, button_type="primary")
            load_data.on_click(lambda event: self.load_data())
            return CenterColumn(load_data,
                                sizing_mode="stretch_both")

    def panels(self):
        return [pn.panel(s) for s in self.sections.values()]
    
    def cards(self):
        cards = [pn.Card(section, header="### "+name, sizing_mode="stretch_both") for name,section in self.sections.items()]
        cards[1].collapsed = True
        return cards

    def _update_login_status(self):
        self._logged_in = self.client.logged_in

    def pre_data_load(self):
        self.loading = True
        if self._cb is not None and self._cb.running:
            self._cb.stop()
        self._loaded = 0
        self._loaded_data_cache = []
        self.data_plotter.loading = True
        self.data_plotter.data = None

    def on_data_load_task_complete(self):
        self._loaded += 1
        logger.info(f"Data loading task {self._loaded} completed.")

    def on_data_load_task_success(self):
        logger.info("Data loading task completed successfully.")

    def on_data_load_success(self):
        self.data_loaded = True
        logger.info("Data loaded successfully.")
        
    def post_data_load(self):
        self.data_plotter.loading = False
        self._loaded_data_cache = []
        if self._cb is not None and self._cb.running:
            self._cb.stop()
        self.loading = False
        logger.info("Data loading completed.")

    def load_data(self):
        if self.loading:
            return "Busy"
        self.pre_data_load()
        try:
            resource = self.filter_tool.get_filtered_resource().paginate(self.partition_size)
            page_numbers = resource.page_numbers
            if not page_numbers:
                self.post_data_load()
                return
            self._npages = len(page_numbers)
            self._ndocs = resource.nitems
            loop = IOLoop.current()
            for idx in page_numbers:
                future = executor.submit(resource.get_page_raw, idx)
                loop.add_future(future, self._data_ready_cb)
            self._cb = pn.state.add_periodic_callback(self._check_data_loading_done, period=1000)
        except Exception as e:
            logger.error("Exception raised while starting data loading jobs: "+str(e))
        
    def _data_ready_cb(self, future):
        try:
            page = future.result()
            if not self.loading:
                return
            with unlocked():
                self._loaded_data_cache.extend(page)
                self.on_data_load_task_success()
        except Exception as e:
                logger.error(str(e))
        finally:
            self.on_data_load_task_complete()
            
    def _check_data_loading_done(self):
        if not self.loading:
            self.post_data_load()
            return

        if self._loaded<self._npages:
            return

        with unlocked():
            try:
                self.process_loaded_data()
                self.on_data_load_success()
            except Exception as e:
                logger.error(str(e))
            finally:
                self.post_data_load()
                
    def process_loaded_data(self):
        df = pd.DataFrame(self._loaded_data_cache)
        df = df[[col for col in df.columns if not col.startswith("_")]]
        if "timestamp" in df.columns and "date" not in df.columns:
            df["date"] = pd.to_datetime(df["timestamp"])
            # .dt.strftime('%Y-%m-%d')
        df = df.replace(-np.inf, np.nan)
        df = df.replace(np.inf, np.nan)
        if self.merge_pmt_info:
            df = pd.merge(df, self.filter_tool.data).dropna(axis=1, how="all").reset_index(drop=True)
        self.data_plotter.data = df
        
    def progress(self):
        nloaded = len(self._loaded_data_cache)
        perc = min(int(100*nloaded/self._ndocs), 100)
        bar = pn.indicators.Progress(value=perc, active=self.loading, max=100,
                                    sizing_mode="stretch_width", margin=10, min_width=300)
        
        text = pn.widgets.StaticText(value=f"Loading requested data. {perc}% [{nloaded}/{self._ndocs} documents] done.", 
                                    align="center")
        indicator = pn.indicators.LoadingSpinner(value=self.loading, width=20, height=20)
        cancel = pn.widgets.Button(name="Cancel")
        def cancel_pressed(event):
            self.loading = False
        cancel.on_click(cancel_pressed)
        return CenterColumn(pn.Row(indicator, text), bar, cancel ,sizing_mode="stretch_width",)

    def data_plot(self):
        if self.data_plotter.loading:
            cancel = pn.widgets.Button(name="Cancel")
            def cancel_pressed(event):
                self.loading = False
            cancel.on_click(cancel_pressed)
            progress = self.progress
            content = pn.Column(
                                progress, 
                                cancel, 
                                sizing_mode="stretch_width")

        elif self.data_plotter.data is None:
            load_data = pn.widgets.Button(name="Load data", 
            height=150, button_type="primary")
            load_data.on_click(lambda event: self.load_data())
            content = CenterColumn(load_data)
        else:
            content = self.data_plotter.view
        return content

    def settings(self):
        load_data = pn.widgets.Button(name="Load data", 
           button_type="primary")
        load_data.on_click(lambda event: self.load_data())
        return pn.Column(
            self.filter_tool.settings,
            self.param.partition_size,
            load_data,
            pn.layout.Divider(),
            self.data_plotter.settings,
            width=270,
            sizing_mode="fixed")


import param
import panel as pn
import holoviews as hv
import numpy as np
import pandas as pd
from pandas.api.types import is_numeric_dtype

from xepmts.web_client.src.shared.templates import ListTemplate
from xepmts.web_client.src.shared import config, get_db
from xepmts.web_client.src.shared._menu import session_menu
from xepmts.web_client.src.shared._session import login_required
from xepmts.web_client.src.shared.models import BaseSection, DataLoader, CenterColumn


class HeatmapPlotter(BaseSection):
    xaxis = param.Selector(["date"], default="date")
    yaxis = param.Selector(["pmt_index"], default="pmt_index")
    color = param.Selector(["gain"], default="gain")
    extra_columns = param.List(["gain_err"])
    catagorical_axes = param.Boolean(default=True)
    groupby = param.Selector(["array"], default="array")
    aggregation = param.Selector(["mean", "median", "min", "max",])
    ncols = param.Integer(2)
    size = param.Integer(12)
    alpha = param.Number(0.9)
    colormap = param.Selector(hv.plotting.list_cmaps(), default="coolwarm")
    
    selection = param.Parameter(precedence=-1)
    _plot = param.Parameter(precedence=-1)
    
    update_plot_action = param.Action(lambda self: self.update_plot(), label="Update Heatmap")

    def _init_view(self):
        self._view = pn.Column("Data not loaded yet.", sizing_mode="stretch_both")
        self._extra_columns = pn.widgets.MultiSelect(options=[], )

    def _update_view(self, *events):
        if self.data is None:
            self.update_plot_action.disabled = True
        else:
            cols = list(self.data.columns)
            self.param.xaxis.objects = cols
            self.param.yaxis.objects = cols
            self.param.color.objects = [col for col in cols if is_numeric_dtype(self.data[col])]
            self._extra_columns.options = cols
            self.update_plot_action.disabled = False
            self.loading = True
            pn.state.add_periodic_callback(self.update_plot, count=1)
            
            
    @param.depends("loading", "_plot")
    def view(self):
        if self.loading:
            return CenterColumn(pn.indicators.LoadingSpinner(value=True),
                                sizing_mode="stretch_both")
            
        if self.data is None or not len(self.data):
            return pn.Column("# No data to show.")
        
        if self._plot is None:
            self.update_plot()

        return pn.Row(self._plot, self.selected_plot())

    def update_plot(self):
        self._plot = self.heatmap_plot()
        self.loading = False

    def heatmap_plot(self):
        if self.xaxis == self.yaxis:
            return pn.Column("X and Y axes are the same.")
        data = self.data.copy()
        if self.catagorical_axes:
            data.sort_values([self.xaxis, self.yaxis])
            for axis in [self.xaxis, self.yaxis]:
                if hasattr(data[axis], "dt"):
                    data[axis] = data[axis].dt.strftime('%Y-%m-%d')
                else:
                    data[axis] = data[axis].astype(str)
            
        aspect = len(self.data[self.xaxis].unique())/len(self.data[self.yaxis].unique())
        vdims = [self.color]+[col for col in self.extra_columns if col not in [self.xaxis, self.yaxis, self.color]]
        hm = hv.HeatMap(data, kdims=[self.xaxis, self.yaxis], vdims=vdims).aggregate(function=getattr(np, self.aggregation))
        hm = hm.opts(tools=["tap", "hover"], nonselection_alpha=0.4, aspect=aspect,
                     responsive=True, alpha=self.alpha, colorbar=True, colorbar_position='top', framewise=True,
                     xrotation=60, xaxis="top", cmap=self.colormap, min_width=600)
        self.selection = hv.streams.Tap(source=hm)
        
        return hm
    
    def selected_plot(self):
        if self.selection is None:
            return pn.Column()
        
        def tap_histogram(x, y):
#             hover = HoverTool(tooltips=[(dim, "@"+dim) for dim in ('cal_bins', "cal_counts")], mode="vline")
            key= (str(x), str(y).split(".")[0])
            # print(key, " Selected.")
            if x is None:
                c = hv.Curve(pd.DataFrame({'bin': [0,1,1,0], "counts": [0.4,0.4,0.5,0.5]}), 'bin', "counts")
                text = hv.Labels(pd.DataFrame({"x":[0.5], "y":[0.5], "text":["No calibration selected\n click on a calibration to view its data."]}), ['x', "y"], "text")
            else:
                counts = np.linspace(-100, 2500, 100)
                random_counts = 100*np.random.random(counts.shape) 
                random_counts += 1000*np.exp(-((counts-np.mean(counts))/2/1000)**2) 
                random_counts += 100*np.random.random(counts.shape) + 1500*np.exp(-(counts/2/50)**2)
                cal_data =  {"bin":counts,
                        "counts": random_counts
                                         }
                c = hv.Curve(pd.DataFrame(cal_data), 'bin', "counts")
                text = hv.Labels(pd.DataFrame({"x":[1000], "y":[1500], "text":["This is random data\n database is still empty"]}), ['x', "y"], "text")                         

            return c.opts(
                hv.opts.Curve(interpolation='steps-mid',framewise=True, height=400,tools=["hover"],
                            line_color='darkred',line_width=1, width=375, yaxis='right', logy=True),
                hv.opts.Labels(height=400, width=375)
                            )*text

        return pn.Column(hv.DynamicMap(tap_histogram, streams=[self.selection]),
                          hv.DynamicMap(tap_histogram, streams=[self.selection]))
    
    def settings(self):
        return pn.Column("### Heatmap options",
                        pn.Param(self.param, expand_button=False,  width=250, show_name=False,
                        parameters=["xaxis", "yaxis", "color", "alpha", "colormap", 
                                    "catagorical_axes", "extra_columns", "update_plot_action"],
                        widgets={"extra_columns": self._extra_columns}
                                 ), width=270, sizing_mode="fixed")
@login_required(pn.state)                    
def view(db):
    # db = pn.state.as_cached(pn.state.curdoc.session_context.id, get_db)
    menu = session_menu(pn.state.curdoc.session_context.id)
    template = ListTemplate(
        title="Heatmap viewer",
        sidebar_footer=menu,
    )
    pn.state.curdoc.theme = template.theme.bokeh_theme

    hm_plotter = HeatmapPlotter()
    loader = DataLoader(client=db, data_plotter=hm_plotter, _logged_in=db.logged_in)

    hv.renderer('bokeh').theme = template.theme.bokeh_theme
    template.main[:] = [pn.panel(c) for c in loader.cards()]
    template.sidebar[:] = [pn.panel(loader.settings)]

    return template


if __name__.startswith("bokeh"):
    view().servable()



from functools import wraps
from xepmts.web_client.src.shared._db import get_db
import panel as pn

def login_required(state, requires_db=True):
    def decorator(view):
        @wraps(view)
        def wrapper():
            db = state.as_cached(state.curdoc.session_context.id, get_db)
            if requires_db:
                template = view(db)
            else:
                template = view()
            if db.logged_in:
                return template
            else:
                message = """
                ## Well, this awkward...

                <p style="text-align: center;">
                 You need to log in to use this app.
                </p>"""
                template.main[:] = [pn.pane.Alert(message)]
                template.sidebar[:] = []
                return template
        return wrapper
    return decorator


import os
from copy import deepcopy


def deep_find_files(path, extensions):
    paths = []
    if not os.path.isdir(path):
        raise ValueError("{} is not a valid directory".format(path))
    for root, dirs, files in os.walk(path):
        for fname in files:
            if fname.split(".")[-1] in extensions:
                paths.append(os.path.join(root, fname))
    return paths

def find_yaml_files(path):
    return deep_find_files(path, ["yaml", "yml"])

def find_json_files(path):
    return deep_find_files(path, ["json"])

def read_endpoint_files(root):
    import yaml
    import json
    import os
    domain = {}
    for path in find_yaml_files(root): 
        with open(path, "r") as f:
            domain.update(yaml.safe_load(f))

    for path in find_json_files(root): 
        with open(path, "r") as f:
            domain.update(json.safe_load(f))
    return domain

def read_endpoint_dirs(roots):
    domain = {}
    for root in roots:
        domain.update(read_endpoint_files(root))

INCLUDE_AS_IS = ["accounts",]
GLOBAL_WRITE_ROLES = ["admin", "expert"]
GLOBAL_READ_ROLES = ["read:all"]

EXPERIMENTS = {
    "xenonnt": {
        "url_prefix": "",
        "name_suffix": "",
        "mongo_prefix": "MONGO",
        "detectors":  ["tpc", "nveto", "muveto"],
    },
    "xenon1t": {
        "url_prefix": "xenon1t",
        "name_suffix": "1t",
        "mongo_prefix": "MONGO",
        "detectors": ["tpc", "muveto"],
    },
}


def resources_from_templates(templates, experiments=EXPERIMENTS,
                             include_as_is=INCLUDE_AS_IS,
                            global_write_roles=GLOBAL_WRITE_ROLES,
                            global_read_roles=GLOBAL_READ_ROLES):
    resources = {}
    for experiment_name, experiment in experiments.items():
        for detector in experiment["detectors"]:
            for name, resource in templates.items():
                if name in include_as_is:
                    resources[name] = resource
                    continue
                new_resource = deepcopy(resource)
                new_resource["item_title"] = "_".join([detector, new_resource["item_title"], experiment["name_suffix"]]).strip("_").title().replace("_", " ")
                new_resource["resource_title"] = "_".join([detector, new_resource["resource_title"], experiment["name_suffix"]]).strip("_").title().replace("_", " ")
                
                new_resource["url"] = '/'.join([experiment["url_prefix"], detector, new_resource.get("url", name)]).strip("/")
                new_resource["datasource"]["source"] = f'{detector}_{new_resource["datasource"]["source"]}'
                if experiment["name_suffix"]:
                    new_resource["datasource"]["source"] += f'_{experiment["name_suffix"]}'
                if "experiment" in new_resource["schema"]:
                    new_resource["schema"]["experiment"]["default"] = experiment_name
                if "detector" in new_resource["schema"]:    
                    new_resource["schema"]["detector"]["default"] = detector
                
                write_roles = global_write_roles + [f'write:{experiment_name}', 
                                             f'write:{experiment_name}_{detector}', 
                                             f'write:{new_resource["datasource"]["source"]}']
                read_roles = write_roles + global_read_roles + [ f'read:{experiment_name}',
                                             f'read:{experiment_name}_{detector}',
                                             f'read:{new_resource["datasource"]["source"]}']
              
                new_resource["allowed_read_roles"] = new_resource.get("allowed_read_roles", []) + read_roles
                new_resource["allowed_item_read_roles"] = new_resource.get("allowed_item_read_roles", []) + read_roles
                new_resource["allowed_write_roles"] = new_resource.get("allowed_write_roles", []) + write_roles
                new_resource["allowed_item_write_roles"] = new_resource.get("allowed_item_write_roles", []) + write_roles
                new_resource["mongo_prefix"] = experiment["mongo_prefix"]
                new_name = "_".join([experiment_name, detector, name])
    
                resources[new_name] = new_resource
                # resources[detector.capitalize() + name + experiment["name_suffix"]] = new_resource
    return resources




"""Provides the MENU html string which is appended to all templates

Please note that the MENU only works in [Fast](https://www.fast.design/) based templates.

If you need some sort of custom MENU html string feel free to customize this code.
"""
from copy import deepcopy
from awesome_panel_extensions.frameworks.fast.fast_menu import to_menu

from xepmts.web_client.src.shared import config

if config.applications:
    MENU = to_menu(
        config.applications.values(), accent_color=config.color_primary, expand=["Main"]
    ).replace("\n", "")
else:
    MENU = ""

def session_menu(session_id):
    applications = []
    for app in config.applications.values():
        session_app = deepcopy(app)
        session_app.url = app.url+f"?bokeh-session-id={session_id}"
        applications.append(session_app)
    menu = to_menu(
        applications, accent_color=config.color_primary, expand=["Main"]
        ).replace("\n", "")
    return menu


"""Contains various modifications, fixes, workarounds and hacks of Panel and other packages to
enable running this site.

Run the `src.modifications.apply` function to apply the modifications"""
import panel as pn
from awesome_panel_extensions.widgets.button import AwesomeButton

from ._server import get_server


def apply():
    """Applies all modifications.

    - Panel fixes
    - Load of awesome-panel-assets .js lib

    Run this before starting the server"""
    # Improvement to be able to pn.serve .py and ipynb files
    pn.io.server.get_server = get_server

    # Hack to get awesome-panel-extensions .js library loaded
    AwesomeButton()



import panel as pn
import holoviews as hv

from src.shared.templates import ListTemplate
from src.shared import config, get_db
from src.shared._menu import session_menu
from src.apps.gains_viewer.gains_viewer import GainsViewer

db = pn.state.as_cached(pn.state.curdoc.session_context.id, get_db)


def view():
    menu = session_menu(pn.state.curdoc.session_context.id)
    template = ListTemplate(
        title="Gain viewer",
        sidebar_footer=menu,

    )
    pn.state.curdoc.theme = template.theme.bokeh_theme
    if db.logged_in:
        gv = GainsViewer(client=db)
        gv._logged_in = db.logged_in 
        pn.state.add_periodic_callback(gv._update_login_status)
        hv.renderer('bokeh').theme = pn.state.curdoc.theme #template.theme.bokeh_theme
        # token_input = 
        template.main[:] = gv.panels()
        # template.header[:] = [token_input]+template.header[:]
        template.sidebar[:] = [gv.settings_view(), pn.Spacer(height=25)]
    else:
        template.main[:] = [pn.pane.Alert("# You need to log in to use this app.")]
    return template


if __name__.startswith("bokeh"):
    view().servable()
